/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 *
 * @format
 */
"use strict";
/* eslint-disable no-bitwise */
// $FlowFixMe: not defined by Flow

function _slicedToArray(arr, i) {
  return (
    _arrayWithHoles(arr) ||
    _iterableToArrayLimit(arr, i) ||
    _unsupportedIterableToArray(arr, i) ||
    _nonIterableRest()
  );
}

function _nonIterableRest() {
  throw new TypeError(
    "Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."
  );
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))
    return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _iterableToArrayLimit(arr, i) {
  if (typeof Symbol === "undefined" || !(Symbol.iterator in Object(arr)))
    return;
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (
      var _i = arr[Symbol.iterator](), _s;
      !(_n = (_s = _i.next()).done);
      _n = true
    ) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

const constants = require("constants");

const stream = require("stream");

const _require = require("events"),
  EventEmitter = _require.EventEmitter;

const kWritableMustExist = Symbol("kWritableMustExist");
const FLAGS_SPECS = {
  r: {
    mustExist: true,
    readable: true
  },
  "r+": {
    mustExist: true,
    readable: true,
    writable: true
  },
  "rs+": {
    mustExist: true,
    readable: true,
    writable: true
  },
  w: {
    truncate: true,
    writable: true
  },
  wx: {
    exclusive: true,
    truncate: true,
    writable: true
  },
  "w+": {
    readable: true,
    truncate: true,
    writable: true
  },
  "wx+": {
    exclusive: true,
    readable: true,
    truncate: true,
    writable: true
  },
  // $FlowIssue: Symbol support is incomplete
  [kWritableMustExist]: {
    mustExist: true,
    writable: true
  }
};
const ASYNC_FUNC_NAMES = [
  "access",
  "chmod",
  "close",
  "copyFile",
  "fchmod",
  "fstat",
  "fsync",
  "fdatasync",
  "lchmod",
  "link",
  "lstat",
  "mkdir",
  "mkdtemp",
  "open",
  "read",
  "readdir",
  "readFile",
  "readlink",
  "realpath",
  "rename",
  "stat",
  "truncate",
  "unlink",
  "write",
  "writeFile"
];
const NATIVE_FUNC_NAMES = ["realpath", "realpathSync"]; // Generated by `node -p "Object.keys(require('fs').promises).sort()"`

const PROMISE_FUNC_NAMES = [
  "access",
  "appendFile",
  "chmod",
  "chown",
  "copyFile",
  "lchmod",
  "lchown",
  "link",
  "lstat",
  "mkdir",
  "mkdtemp",
  "open",
  "readFile",
  "readdir",
  "readlink",
  "realpath",
  "rename",
  "rmdir",
  "stat",
  "symlink",
  "truncate",
  "unlink",
  "utimes",
  "writeFile"
];
/**
 * Simulates `fs` API in an isolated, memory-based filesystem. This is useful
 * for testing systems that rely on `fs` without affecting the real filesystem.
 * This is meant to be a drop-in replacement/mock for `fs`, so it mimics
 * closely the behavior of file path resolution and file accesses.
 */

class MemoryFs {
  constructor(_options) {
    var _this = this;

    _defineProperty(this, "constants", constants);

    _defineProperty(this, "Dirent", Dirent);

    _defineProperty(this, "accessSync", (filePath, mode) => {
      if (mode == null) {
        mode = constants.F_OK;
      }

      const stats = this.statSync(filePath);

      if (mode == constants.F_OK) {
        return;
      }

      const filePathStr = pathStr(filePath);

      if ((mode & constants.R_OK) !== 0) {
        if (
          !(
            (stats.mode & constants.S_IROTH) !== 0 ||
            ((stats.mode & constants.S_IRGRP) !== 0 &&
              stats.gid === getgid()) ||
            ((stats.mode & constants.S_IRUSR) !== 0 && stats.uid === getuid())
          )
        ) {
          throw makeError("EPERM", filePathStr, "file cannot be read");
        }
      }

      if ((mode & constants.W_OK) !== 0) {
        if (
          !(
            (stats.mode & constants.S_IWOTH) !== 0 ||
            ((stats.mode & constants.S_IWGRP) !== 0 &&
              stats.gid === getgid()) ||
            ((stats.mode & constants.S_IWUSR) !== 0 && stats.uid === getuid())
          )
        ) {
          throw makeError("EPERM", filePathStr, "file cannot be written to");
        }
      }

      if ((mode & constants.X_OK) !== 0) {
        if (
          !(
            (stats.mode & constants.S_IXOTH) !== 0 ||
            ((stats.mode & constants.S_IXGRP) !== 0 &&
              stats.gid === getgid()) ||
            ((stats.mode & constants.S_IXUSR) !== 0 && stats.uid === getuid())
          )
        ) {
          throw makeError("EPERM", filePathStr, "file cannot be executed");
        }
      }
    });

    _defineProperty(this, "chmodSync", (filePath, mode) => {
      filePath = pathStr(filePath);
      mode = parseFileMode(mode);

      const _this$_resolve = this._resolve(filePath),
        node = _this$_resolve.node;

      if (node == null) {
        throw makeError("ENOENT", filePath, "no such file or directory");
      }

      node.mode = mode;
    });

    _defineProperty(this, "closeSync", fd => {
      const desc = this._getDesc(fd);

      if (desc.writable) {
        this._emitFileChange(desc.nodePath.slice(), {
          eventType: "change"
        });
      }

      this._fds.delete(fd);
    });

    _defineProperty(this, "copyFileSync", function(src, dest) {
      let flags =
        arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;
      const options =
        flags & constants.COPYFILE_EXCL
          ? {
              flag: "wx"
            }
          : {};

      _this.writeFileSync(dest, _this.readFileSync(src), options);
    });

    _defineProperty(this, "fchmodSync", (fd, mode) => {
      mode = parseFileMode(mode);

      const _this$_getDesc = this._getDesc(fd),
        node = _this$_getDesc.node;

      node.mode = mode;
    });

    _defineProperty(this, "fsyncSync", fd => {
      this._getDesc(fd);
    });

    _defineProperty(this, "fdatasyncSync", fd => {
      this._getDesc(fd);
    });

    _defineProperty(this, "lchmodSync", (filePath, mode) => {
      filePath = pathStr(filePath);
      mode = parseFileMode(mode);

      const _this$_resolve2 = this._resolve(filePath, {
          keepFinalSymlink: true
        }),
        node = _this$_resolve2.node;

      if (node == null) {
        throw makeError("ENOENT", filePath, "no such file or directory");
      }

      node.mode = mode;
    });

    _defineProperty(this, "openSync", (filePath, flags, mode) => {
      if (typeof flags === "number") {
        throw new Error(`numeric flags not supported: ${flags}`);
      }

      return this._open(pathStr(filePath), flags, mode);
    });

    _defineProperty(
      this,
      "readSync",
      (fd, buffer, offset, length, position) => {
        const desc = this._getDesc(fd);

        if (!desc.readable) {
          throw makeError(
            "EBADF",
            null,
            "file descriptor cannot be written to"
          );
        }

        if (position != null) {
          desc.position = position;
        }

        const endPos = Math.min(
          desc.position + length,
          desc.node.content.length
        );
        desc.node.content.copy(buffer, offset, desc.position, endPos);
        const bytesRead = endPos - desc.position;
        desc.position = endPos;
        return bytesRead;
      }
    );

    _defineProperty(this, "readdirSync", (filePath, options) => {
      let encoding, withFileTypes;

      if (typeof options === "string") {
        encoding = options;
      } else if (options != null) {
        encoding = options.encoding;
        withFileTypes = options.withFileTypes;
      }

      filePath = pathStr(filePath);

      const _this$_resolve3 = this._resolve(filePath),
        node = _this$_resolve3.node;

      if (node == null) {
        throw makeError("ENOENT", filePath, "no such file or directory");
      }

      if (node.type !== "directory") {
        throw makeError("ENOTDIR", filePath, "not a directory");
      }

      return Array.from(node.entries.keys()).map(str => {
        let name;

        if (encoding === "utf8") {
          name = str;
        } else {
          const buffer = Buffer.from(str);

          if (encoding === "buffer") {
            name = buffer;
          } else {
            name = buffer.toString(encoding);
          }
        }

        if (withFileTypes) {
          return new Dirent(nullthrows(node.entries.get(str)), name);
        }

        return name;
      });
    });

    _defineProperty(this, "readFileSync", (filePath, options) => {
      let encoding, flag;

      if (typeof options === "string") {
        encoding = options;
      } else if (options != null) {
        encoding = options.encoding;
        flag = options.flag;
      }

      const fd = this._open(pathStr(filePath), flag || "r");

      const chunks = [];

      try {
        const buffer = Buffer.alloc(1024);
        let bytesRead;

        do {
          bytesRead = this.readSync(fd, buffer, 0, buffer.length, null);

          if (bytesRead === 0) {
            continue;
          }

          const chunk = Buffer.alloc(bytesRead);
          buffer.copy(chunk, 0, 0, bytesRead);
          chunks.push(chunk);
        } while (bytesRead > 0);
      } finally {
        this.closeSync(fd);
      }

      const result = Buffer.concat(chunks);

      if (encoding == null) {
        return result;
      }

      return result.toString(encoding);
    });

    _defineProperty(this, "readlinkSync", (filePath, options) => {
      let encoding;

      if (typeof options === "string") {
        encoding = options;
      } else if (options != null) {
        encoding = options.encoding;
      }

      filePath = pathStr(filePath);

      const _this$_resolve4 = this._resolve(filePath, {
          keepFinalSymlink: true
        }),
        node = _this$_resolve4.node;

      if (node == null) {
        throw makeError("ENOENT", filePath, "no such file or directory");
      }

      if (node.type !== "symbolicLink") {
        throw makeError("EINVAL", filePath, "entity is not a symlink");
      }

      if (encoding == null || encoding === "utf8") {
        return node.target;
      }

      const buf = Buffer.from(node.target);

      if (encoding == "buffer") {
        return buf;
      }

      return buf.toString(encoding);
    });

    _defineProperty(this, "realpathSync", filePath => {
      return this._resolve(pathStr(filePath)).realpath;
    });

    _defineProperty(
      this,
      "writeSync",
      (fd, bufferOrString, offsetOrPosition, lengthOrEncoding, position) => {
        let encoding, offset, length, buffer;

        if (typeof bufferOrString === "string") {
          position = offsetOrPosition;
          encoding = lengthOrEncoding;
          buffer = Buffer.from(bufferOrString, encoding || "utf8");
        } else {
          offset = offsetOrPosition;

          if (
            lengthOrEncoding != null &&
            typeof lengthOrEncoding !== "number"
          ) {
            throw new Error("invalid length");
          }

          length = lengthOrEncoding;
          buffer = bufferOrString;
        }

        if (offset == null) {
          offset = 0;
        }

        if (length == null) {
          length = buffer.length;
        }

        return this._write(fd, buffer, offset, length, position);
      }
    );

    _defineProperty(this, "writeFileSync", (filePathOrFd, data, options) => {
      let encoding, mode, flag;

      if (typeof options === "string") {
        encoding = options;
      } else if (options != null) {
        encoding = options.encoding;
        mode = options.mode;
        flag = options.flag;
      }

      if (encoding == null) {
        encoding = "utf8";
      }

      if (typeof data === "string") {
        data = Buffer.from(data, encoding);
      }

      const fd =
        typeof filePathOrFd === "number"
          ? filePathOrFd
          : this._open(pathStr(filePathOrFd), flag || "w", mode);

      try {
        this._write(fd, data, 0, data.length);
      } finally {
        if (typeof filePathOrFd !== "number") {
          this.closeSync(fd);
        }
      }
    });

    _defineProperty(this, "mkdirSync", (dirPath, mode) => {
      if (mode == null) {
        mode = 0o777;
      }

      dirPath = pathStr(dirPath);

      const _this$_resolve5 = this._resolve(dirPath),
        dirNode = _this$_resolve5.dirNode,
        node = _this$_resolve5.node,
        basename = _this$_resolve5.basename;

      if (node != null) {
        throw makeError("EEXIST", dirPath, "directory or file already exists");
      }

      dirNode.entries.set(basename, this._makeDir(mode));
    });

    _defineProperty(this, "mkdtempSync", (prefix, options) => {
      var _ref;

      const encoding =
        (_ref =
          typeof options === "string"
            ? options
            : options === null || options === void 0
            ? void 0
            : options.encoding) !== null && _ref !== void 0
          ? _ref
          : "utf8";
      const ALPHABET =
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_";
      const RANDOM_PART_LENGTH = 6;
      let dirPath = "";

      do {
        let randomPart = "";

        for (let i = 0; i < RANDOM_PART_LENGTH; ++i) {
          randomPart += ALPHABET[Math.floor(Math.random() * ALPHABET.length)];
        }

        dirPath = prefix + randomPart;
      } while (this.existsSync(dirPath));

      this.mkdirSync(dirPath, 0o700);
      return Buffer.from(dirPath, "utf8").toString(encoding);
    });

    _defineProperty(this, "rmdirSync", dirPath => {
      dirPath = pathStr(dirPath);

      const _this$_resolve6 = this._resolve(dirPath),
        dirNode = _this$_resolve6.dirNode,
        node = _this$_resolve6.node,
        basename = _this$_resolve6.basename;

      if (node == null) {
        throw makeError("ENOENT", dirPath, "directory does not exist");
      } else if (node.type === "file") {
        if (this._platform === "posix") {
          throw makeError("ENOTDIR", dirPath, "cannot rm a file");
        } else {
          throw makeError("ENOENT", dirPath, "cannot rm a file");
        }
      } else if (node.type === "directory" && node.entries.size) {
        throw makeError("ENOTEMPTY", dirPath, "directory not empty");
      }

      dirNode.entries.delete(basename);
    });

    _defineProperty(this, "symlinkSync", (target, filePath, type) => {
      if (type == null) {
        type = "file";
      }

      if (type !== "file") {
        throw new Error("symlink type not supported");
      }

      filePath = pathStr(filePath);

      const _this$_resolve7 = this._resolve(filePath),
        dirNode = _this$_resolve7.dirNode,
        node = _this$_resolve7.node,
        basename = _this$_resolve7.basename;

      if (node != null) {
        throw makeError("EEXIST", filePath, "directory or file already exists");
      }

      dirNode.entries.set(basename, {
        id: this._getId(),
        gid: getgid(),
        target: pathStr(target),
        mode: 0o666,
        uid: getuid(),
        type: "symbolicLink",
        watchers: []
      });
    });

    _defineProperty(this, "existsSync", filePath => {
      try {
        const _this$_resolve8 = this._resolve(pathStr(filePath)),
          node = _this$_resolve8.node;

        return node != null;
      } catch (error) {
        if (error.code === "ENOENT") {
          return false;
        }

        throw error;
      }
    });

    _defineProperty(this, "statSync", filePath => {
      filePath = pathStr(filePath);

      const _this$_resolve9 = this._resolve(filePath),
        node = _this$_resolve9.node;

      if (node == null) {
        throw makeError("ENOENT", filePath, "no such file or directory");
      }

      return new Stats(node);
    });

    _defineProperty(this, "lstatSync", filePath => {
      filePath = pathStr(filePath);

      const _this$_resolve10 = this._resolve(filePath, {
          keepFinalSymlink: true
        }),
        node = _this$_resolve10.node;

      if (node == null) {
        throw makeError("ENOENT", filePath, "no such file or directory");
      }

      return new Stats(node);
    });

    _defineProperty(this, "fstatSync", fd => {
      const desc = this._getDesc(fd);

      return new Stats(desc.node);
    });

    _defineProperty(this, "createReadStream", (filePath, options) => {
      let autoClose, encoding, fd, flags, mode, start, end, highWaterMark;

      if (typeof options === "string") {
        encoding = options;
      } else if (options != null) {
        autoClose = options.autoClose;
        encoding = options.encoding;
        fd = options.fd;
        flags = options.flags;
        mode = options.mode;
        start = options.start;
        end = options.end;
        highWaterMark = options.highWaterMark;
      }

      let st = null;

      if (fd == null) {
        fd = this._open(pathStr(filePath), flags || "r", mode);
        process.nextTick(() => st.emit("open", fd));
      }

      const ffd = fd;
      const readSync = this.readSync;
      const ropt = {
        filePath,
        encoding,
        fd,
        highWaterMark,
        start,
        end,
        readSync
      };
      const rst = new ReadFileSteam(ropt);
      st = rst;

      if (autoClose !== false) {
        const doClose = () => {
          this.closeSync(ffd);
          rst.emit("close");
        };

        rst.on("end", doClose);
        rst.on("error", doClose);
      }

      return rst;
    });

    _defineProperty(this, "unlinkSync", filePath => {
      filePath = pathStr(filePath);

      const _this$_resolve11 = this._resolve(filePath, {
          keepFinalSymlink: true
        }),
        basename = _this$_resolve11.basename,
        dirNode = _this$_resolve11.dirNode,
        dirPath = _this$_resolve11.dirPath,
        node = _this$_resolve11.node;

      if (node == null) {
        throw makeError("ENOENT", filePath, "no such file or directory");
      }

      if (node.type !== "file" && node.type !== "symbolicLink") {
        throw makeError("EISDIR", filePath, "cannot unlink a directory");
      }

      dirNode.entries.delete(basename);

      this._emitFileChange(dirPath.concat([[basename, node]]), {
        eventType: "rename"
      });
    });

    _defineProperty(this, "renameSync", (oldPath, newPath) => {
      oldPath = pathStr(oldPath);
      newPath = pathStr(newPath);

      const _this$_resolve12 = this._resolve(oldPath, {
          keepFinalSymlink: true
        }),
        oldBasename = _this$_resolve12.basename,
        oldDirNode = _this$_resolve12.dirNode,
        oldDirPath = _this$_resolve12.dirPath,
        node = _this$_resolve12.node;

      if (node == null) {
        throw makeError("ENOENT", oldPath, "no such file or directory");
      }

      const _this$_resolve13 = this._resolve(newPath, {
          keepFinalSymlink: true
        }),
        newBasename = _this$_resolve13.basename,
        newDirNode = _this$_resolve13.dirNode,
        newDirPath = _this$_resolve13.dirPath,
        existingDestNode = _this$_resolve13.node;

      if (existingDestNode === node) {
        return;
      }

      if (
        newDirPath.some(_ref2 => {
          let _ref3 = _slicedToArray(_ref2, 2),
            nodeInNewPath = _ref3[1];

          return nodeInNewPath === node;
        })
      ) {
        throw makeError(
          "EINVAL",
          newPath,
          "cannot make a directory a subdirectory of itself"
        );
      }

      if (existingDestNode) {
        if (existingDestNode.type === "directory") {
          if (existingDestNode.entries.size) {
            throw makeError("ENOTEMPTY", newPath, "directory not empty");
          }
        } else if (node.type === "directory") {
          throw makeError(
            "EISDIR",
            newPath,
            "cannot overwrite a directory with a non-directory"
          );
        }
      }

      newDirNode.entries.set(newBasename, node);

      if (existingDestNode) {
        // The existing node has been removed.
        this._emitFileChange(
          newDirPath.concat([[newBasename, existingDestNode]]),
          {
            eventType: "rename"
          }
        );
      } // The source node has been linked at the new path.

      this._emitFileChange(newDirPath.concat([[newBasename, node]]), {
        eventType: "rename"
      });

      oldDirNode.entries.delete(oldBasename); // The source node has been unlinked at the old path.

      this._emitFileChange(oldDirPath.concat([[oldBasename, node]]), {
        eventType: "rename"
      });
    });

    _defineProperty(this, "linkSync", (oldPath, newPath) => {
      oldPath = pathStr(oldPath);
      newPath = pathStr(newPath);

      const _this$_resolve14 = this._resolve(oldPath),
        node = _this$_resolve14.node;

      if (node == null) {
        throw makeError("ENOENT", oldPath, "no such file or directory");
      }

      if (node.type === "directory") {
        throw makeError(
          "EPERM",
          oldPath,
          "cannot create a hard link to a directory"
        );
      }

      const _this$_resolve15 = this._resolve(newPath),
        newBasename = _this$_resolve15.basename,
        newDirNode = _this$_resolve15.dirNode,
        newDirPath = _this$_resolve15.dirPath,
        existingDestNode = _this$_resolve15.node;

      if (existingDestNode) {
        throw makeError("EEXIST", newPath, "destination path already exists");
      }

      newDirNode.entries.set(newBasename, node);

      this._emitFileChange(newDirPath.concat([[newBasename, node]]), {
        eventType: "rename"
      });
    });

    _defineProperty(this, "truncateSync", function(filePathOrFd) {
      let length =
        arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
      const fd =
        typeof filePathOrFd === "number"
          ? filePathOrFd
          : _this._open(pathStr(filePathOrFd), kWritableMustExist);

      try {
        const desc = _this._getDesc(fd);

        if (!desc.writable) {
          throw makeError(
            "EBADF",
            null,
            "file descriptor cannot be written to"
          );
        }

        const node = desc.node,
          nodePath = desc.nodePath;
        const oldContent = node.content;
        node.content = Buffer.alloc(length, 0);
        oldContent.copy(
          node.content,
          0,
          0,
          Math.max(0, Math.min(length, oldContent.length))
        );

        _this._emitFileChange(nodePath.slice(), {
          eventType: "change"
        });
      } finally {
        if (typeof filePathOrFd !== "number") {
          _this.closeSync(fd);
        }
      }
    });

    _defineProperty(this, "createWriteStream", (filePath, options) => {
      let autoClose, fd, flags, mode, start;

      if (typeof options !== "string" && options != null) {
        autoClose = options.autoClose;
        fd = options.fd;
        flags = options.flags;
        mode = options.mode;
        start = options.start;
      }

      let st = null;

      if (fd == null) {
        fd = this._open(pathStr(filePath), flags || "w", mode);
        process.nextTick(() => st.emit("open", fd));
      }

      const ffd = fd;
      const ropt = {
        fd,
        writeSync: this._write.bind(this),
        filePath,
        start
      };
      const rst = new WriteFileStream(ropt);
      st = rst;

      if (autoClose !== false) {
        const doClose = () => {
          this.closeSync(ffd);
          rst.emit("close");
        };

        rst.on("finish", doClose);
        rst.on("error", doClose);
      }

      return st;
    });

    _defineProperty(this, "watch", (filePath, options, listener) => {
      filePath = pathStr(filePath);

      const _this$_resolve16 = this._resolve(filePath),
        node = _this$_resolve16.node;

      if (node == null) {
        throw makeError("ENOENT", filePath, "no such file or directory");
      }

      let encoding, recursive, persistent;

      if (typeof options === "string") {
        encoding = options;
      } else if (options != null) {
        encoding = options.encoding;
        recursive = options.recursive;
        persistent = options.persistent;
      }

      const watcher = new FSWatcher(node, {
        encoding: encoding != null ? encoding : "utf8",
        recursive: recursive != null ? recursive : false,
        persistent: persistent != null ? persistent : false
      });

      if (listener != null) {
        watcher.on("change", listener);
      }

      return watcher;
    });

    this._platform = (_options && _options.platform) || "posix";
    this._cwd = _options && _options.cwd;
    this._pathSep = this._platform === "win32" ? "\\" : "/";
    this.reset();
    ASYNC_FUNC_NAMES.forEach(funcName => {
      const func = this[`${funcName}Sync`];

      this[funcName] = function() {
        for (
          var _len = arguments.length, args = new Array(_len), _key = 0;
          _key < _len;
          _key++
        ) {
          args[_key] = arguments[_key];
        }

        const callback = args.pop();
        process.nextTick(() => {
          let retval;

          try {
            retval = func.apply(null, args);
          } catch (error) {
            callback(error);
            return;
          }

          callback(null, retval);
        });
      };
    });
    NATIVE_FUNC_NAMES.forEach(funcName => {
      const func = this[funcName];
      func.native = func;
    });
    this.promises = PROMISE_FUNC_NAMES.filter(
      // $FlowFixMe: No indexer
      funcName => typeof this[`${funcName}Sync`] === "function"
    ).reduce((promises, funcName) => {
      promises[funcName] = function() {
        for (
          var _len2 = arguments.length, args = new Array(_len2), _key2 = 0;
          _key2 < _len2;
          _key2++
        ) {
          args[_key2] = arguments[_key2];
        }

        return new Promise((resolve, reject) => {
          try {
            // $FlowFixMe: No indexer
            resolve(_this[`${funcName}Sync`].apply(_this, args));
          } catch (error) {
            reject(error);
          }
        });
      };

      return promises;
    }, {});
  }

  reset() {
    this._nextId = 1;
    this._roots = new Map();

    if (this._platform === "posix") {
      this._roots.set("", this._makeDir(0o777));
    } else if (this._platform === "win32") {
      this._roots.set("C:", this._makeDir(0o777));
    }

    this._fds = new Map();
  }

  _makeDir(mode) {
    return {
      entries: new Map(),
      gid: getgid(),
      id: this._getId(),
      mode,
      uid: getuid(),
      type: "directory",
      watchers: []
    };
  }

  _getId() {
    return ++this._nextId;
  }

  _open(filePath, flags, mode) {
    if (mode == null) {
      mode = 0o666;
    }

    const spec = FLAGS_SPECS[flags];

    if (spec == null) {
      throw new Error(`flags not supported: \`${flags.toString()}\``);
    }

    const _spec$writable = spec.writable,
      writable = _spec$writable === void 0 ? false : _spec$writable,
      _spec$readable = spec.readable,
      readable = _spec$readable === void 0 ? false : _spec$readable;
    const exclusive = spec.exclusive,
      mustExist = spec.mustExist,
      truncate = spec.truncate;

    let _this$_resolve17 = this._resolve(filePath),
      dirNode = _this$_resolve17.dirNode,
      node = _this$_resolve17.node,
      basename = _this$_resolve17.basename,
      dirPath = _this$_resolve17.dirPath;

    let nodePath;

    if (node == null) {
      if (mustExist) {
        throw makeError("ENOENT", filePath, "no such file or directory");
      }

      node = {
        content: Buffer.alloc(0),
        gid: getgid(),
        id: this._getId(),
        mode,
        uid: getuid(),
        type: "file",
        watchers: []
      };
      dirNode.entries.set(basename, node);
      nodePath = dirPath.concat([[basename, node]]);

      this._emitFileChange(nodePath.slice(), {
        eventType: "rename"
      });
    } else {
      if (exclusive) {
        throw makeError("EEXIST", filePath, "directory or file already exists");
      }

      if (node.type !== "file") {
        throw makeError("EISDIR", filePath, "cannot read/write to a directory");
      }

      if (truncate) {
        node.content = Buffer.alloc(0);
      }

      nodePath = dirPath.concat([[basename, node]]);
    }

    return this._getFd(filePath, {
      nodePath,
      node,
      position: 0,
      readable,
      writable
    });
  }

  _parsePath(filePath) {
    let drive;
    const sep = this._platform === "win32" ? /[\\/]/ : /\//;

    if (this._platform === "win32" && filePath.match(/^[a-zA-Z]:[\\/]/)) {
      drive = filePath.substring(0, 2);
      filePath = filePath.substring(3);
    }

    if (sep.test(filePath[0])) {
      if (this._platform === "posix") {
        drive = "";
        filePath = filePath.substring(1);
      } else {
        throw makeError(
          "EINVAL",
          filePath,
          "path is invalid because it cannot start with a separator"
        );
      }
    }

    return {
      entNames: filePath.split(sep),
      drive
    };
  }
  /**
   * Implemented according with
   * http://man7.org/linux/man-pages/man7/path_resolution.7.html
   */

  _resolve(filePath, options) {
    let keepFinalSymlink = false;

    if (options != null) {
      keepFinalSymlink = options.keepFinalSymlink;
    }

    if (filePath === "") {
      throw makeError("ENOENT", filePath, "no such file or directory");
    }

    let _this$_parsePath = this._parsePath(filePath),
      drive = _this$_parsePath.drive,
      entNames = _this$_parsePath.entNames;

    if (drive == null) {
      const _cwd = this._cwd;

      if (_cwd == null) {
        throw new Error(
          `The path \`${filePath}\` cannot be resolved because no ` +
            "current working directory function has been specified. Set the " +
            "`cwd` option field to specify a current working directory."
        );
      }

      const cwPath = this._parsePath(_cwd());

      drive = cwPath.drive;

      if (drive == null) {
        throw new Error(
          "On a win32 FS, the options' `cwd()` must return a valid win32 " +
            "absolute path. This happened while trying to " +
            `resolve: \`${filePath}\``
        );
      }

      entNames = cwPath.entNames.concat(entNames);
    }

    checkPathLength(entNames, filePath);

    const root = this._getRoot(drive, filePath);

    const context = {
      drive,
      node: root,
      nodePath: [["", root]],
      entNames,
      symlinkCount: 0,
      keepFinalSymlink
    };

    while (context.entNames.length > 0) {
      const entName = context.entNames.shift();

      this._resolveEnt(context, filePath, entName);
    }

    const nodePath = context.nodePath;
    return {
      drive: context.drive,
      realpath: context.drive + nodePath.map(x => x[0]).join(this._pathSep),
      dirNode: (() => {
        const dirNode =
          nodePath.length >= 2
            ? nodePath[nodePath.length - 2][1]
            : context.node;

        if (dirNode == null || dirNode.type !== "directory") {
          throw new Error("failed to resolve");
        }

        return dirNode;
      })(),
      node: context.node,
      basename: nullthrows(nodePath[nodePath.length - 1][0]),
      dirPath: nodePath
        .slice(0, -1)
        .map(nodePair => [nodePair[0], nullthrows(nodePair[1])])
    };
  }

  _resolveEnt(context, filePath, entName) {
    const node = context.node;

    if (node == null) {
      throw makeError("ENOENT", filePath, "no such file or directory");
    }

    if (node.type !== "directory") {
      throw makeError("ENOTDIR", filePath, "not a directory");
    }

    const entries = node.entries;

    if (entName === "" || entName === ".") {
      return;
    }

    if (entName === "..") {
      const nodePath = context.nodePath;

      if (nodePath.length > 1) {
        nodePath.pop();
        context.node = nodePath[nodePath.length - 1][1];
      }

      return;
    }

    const childNode = entries.get(entName);

    if (
      childNode == null ||
      childNode.type !== "symbolicLink" ||
      (context.keepFinalSymlink && context.entNames.length === 0)
    ) {
      context.node = childNode;
      context.nodePath.push([entName, childNode]);
      return;
    }

    if (context.symlinkCount >= 10) {
      throw makeError("ELOOP", filePath, "too many levels of symbolic links");
    }

    const _this$_parsePath2 = this._parsePath(childNode.target),
      entNames = _this$_parsePath2.entNames,
      drive = _this$_parsePath2.drive;

    if (drive != null) {
      context.drive = drive;
      context.node = this._getRoot(drive, filePath);
      context.nodePath = [["", context.node]];
    }

    context.entNames = entNames.concat(context.entNames);
    checkPathLength(context.entNames, filePath);
    ++context.symlinkCount;
  }

  _getRoot(drive, filePath) {
    const root = this._roots.get(drive.toUpperCase());

    if (root == null) {
      throw makeError("ENOENT", filePath, `no such drive: \`${drive}\``);
    }

    return root;
  }

  _write(fd, buffer, offset, length, position) {
    const desc = this._getDesc(fd);

    if (!desc.writable) {
      throw makeError("EBADF", null, "file descriptor cannot be written to");
    }

    if (position == null) {
      position = desc.position;
    }

    const node = desc.node;

    if (node.content.length < position + length) {
      const newBuffer = Buffer.alloc(position + length);
      node.content.copy(newBuffer, 0, 0, node.content.length);
      node.content = newBuffer;
    }

    buffer.copy(node.content, position, offset, offset + length);
    desc.position = position + length;
    return buffer.length;
  }

  _getFd(filePath, desc) {
    let fd = 3;

    while (this._fds.has(fd)) {
      ++fd;
    }

    if (fd >= 256) {
      throw makeError("EMFILE", filePath, "too many open files");
    }

    this._fds.set(fd, desc);

    return fd;
  }

  _getDesc(fd) {
    const desc = this._fds.get(fd);

    if (desc == null) {
      throw makeError("EBADF", null, "file descriptor is not open");
    }

    return desc;
  }

  _emitFileChange(nodePath, options) {
    const fileNode = nodePath.pop();
    let filePath = fileNode[0];
    let recursive = false;

    for (const watcher of fileNode[1].watchers) {
      watcher.listener(options.eventType, filePath);
    }

    while (nodePath.length > 0) {
      const dirNode = nodePath.pop();

      for (const watcher of dirNode[1].watchers) {
        if (recursive && !watcher.recursive) {
          continue;
        }

        watcher.listener(options.eventType, filePath);
      }

      filePath = dirNode[0] + this._pathSep + filePath;
      recursive = true;
    }
  }
}

class Stats {
  /**
   * Don't keep a reference to the node as it may get mutated over time.
   */
  constructor(node) {
    this._type = node.type;
    this.dev = 1;
    this.mode = node.mode;
    this.nlink = 1;
    this.uid = node.uid;
    this.gid = node.gid;
    this.rdev = 0;
    this.blksize = 1024;
    this.ino = node.id;
    this.size =
      node.type === "file"
        ? node.content.length
        : node.type === "symbolicLink"
        ? node.target.length
        : 0;
    this.blocks = Math.ceil(this.size / 512);
    this.atimeMs = 1;
    this.mtimeMs = 1;
    this.ctimeMs = 1;
    this.birthtimeMs = 1;
    this.atime = new Date(this.atimeMs);
    this.mtime = new Date(this.mtimeMs);
    this.ctime = new Date(this.ctimeMs);
    this.birthtime = new Date(this.birthtimeMs);
  }

  isFile() {
    return this._type === "file";
  }

  isDirectory() {
    return this._type === "directory";
  }

  isBlockDevice() {
    return false;
  }

  isCharacterDevice() {
    return false;
  }

  isSymbolicLink() {
    return this._type === "symbolicLink";
  }

  isFIFO() {
    return false;
  }

  isSocket() {
    return false;
  }
}

class ReadFileSteam extends stream.Readable {
  constructor(options) {
    const highWaterMark = options.highWaterMark,
      fd = options.fd;
    const superOptions =
      highWaterMark != null
        ? {
            highWaterMark
          }
        : {};
    super(superOptions);
    this.bytesRead = 0;
    this.path = options.filePath;
    this._readSync = options.readSync;
    this._fd = fd;
    this._buffer = Buffer.alloc(1024);
    const start = options.start,
      end = options.end;

    if (start != null) {
      this._readSync(fd, Buffer.alloc(0), 0, 0, start);
    }

    if (end != null) {
      this._positions = {
        current: start || 0,
        last: end + 1
      };
    }
  }

  _read(size) {
    let bytesRead;
    const _buffer = this._buffer;

    do {
      const length = this._getLengthToRead();

      const position = this._positions && this._positions.current;
      bytesRead = this._readSync(this._fd, _buffer, 0, length, position);

      if (this._positions != null) {
        this._positions.current += bytesRead;
      }

      this.bytesRead += bytesRead;
    } while (this.push(bytesRead > 0 ? _buffer.slice(0, bytesRead) : null));
  }

  _getLengthToRead() {
    const _positions = this._positions,
      _buffer = this._buffer;

    if (_positions == null) {
      return _buffer.length;
    }

    const leftToRead = Math.max(0, _positions.last - _positions.current);
    return Math.min(_buffer.length, leftToRead);
  }
}

class WriteFileStream extends stream.Writable {
  constructor(opts) {
    super();
    this.path = opts.filePath;
    this.bytesWritten = 0;
    this._fd = opts.fd;
    this._writeSync = opts.writeSync;

    if (opts.start != null) {
      this._writeSync(opts.fd, Buffer.alloc(0), 0, 0, opts.start);
    }
  }

  _write(buffer, encoding, callback) {
    try {
      const bytesWritten = this._writeSync(this._fd, buffer, 0, buffer.length);

      this.bytesWritten += bytesWritten;
    } catch (error) {
      callback(error);
      return;
    }

    callback();
  }
}

class FSWatcher extends EventEmitter {
  constructor(node, options) {
    super();

    _defineProperty(this, "_listener", (eventType, filePath) => {
      const encFilePath =
        this._encoding === "buffer" ? Buffer.from(filePath, "utf8") : filePath;

      try {
        this.emit("change", eventType, encFilePath);
      } catch (error) {
        this.close();
        this.emit("error", error);
      }
    });

    this._encoding = options.encoding;
    this._nodeWatcher = {
      recursive: options.recursive,
      listener: this._listener
    };
    node.watchers.push(this._nodeWatcher);
    this._node = node;

    if (options.persistent) {
      this._persistIntervalId = setInterval(() => {}, 60000);
    }
  }

  close() {
    this._node.watchers.splice(this._node.watchers.indexOf(this._nodeWatcher));

    clearInterval(this._persistIntervalId);
  }
}

class Dirent {
  /**
   * Don't keep a reference to the node as it may get mutated over time.
   */
  constructor(node, name) {
    this._stats = new Stats(node);
    this.name = name;
  }

  isBlockDevice() {
    return this._stats.isBlockDevice();
  }

  isCharacterDevice() {
    return this._stats.isCharacterDevice();
  }

  isDirectory() {
    return this._stats.isDirectory();
  }

  isFIFO() {
    return this._stats.isFIFO();
  }

  isFile() {
    return this._stats.isFile();
  }

  isSocket() {
    return this._stats.isSocket();
  }

  isSymbolicLink() {
    return this._stats.isSymbolicLink();
  }
}

function checkPathLength(entNames, filePath) {
  if (entNames.length > 32) {
    throw makeError(
      "ENAMETOOLONG",
      filePath,
      "file path too long (or one of the intermediate " +
        "symbolic link resolutions)"
    );
  }
}

function pathStr(filePath) {
  if (typeof filePath === "string") {
    return filePath;
  }

  return filePath.toString("utf8");
}

function makeError(code, filePath, message) {
  const err = new Error(
    filePath != null
      ? `${code}: \`${filePath}\`: ${message}`
      : `${code}: ${message}`
  );
  err.code = code;
  err.errno = constants[code];
  err.path = filePath;
  return err;
}

function nullthrows(x) {
  if (x == null) {
    throw new Error("item was null or undefined");
  }

  return x;
}

function getgid() {
  return process.getgid != null ? process.getgid() : -1;
}

function getuid() {
  return process.getuid != null ? process.getuid() : -1;
}

function parseFileMode(mode) {
  if (typeof mode === "string") {
    return Number.parseInt(mode, 8);
  }

  return mode;
}

module.exports = MemoryFs;
